/****************************************************************************
*
*   Since this code originated from code which is public domain, I
*   hereby declare this code to be public domain as well.
*
****************************************************************************/
/**
*
*   @file   CBUF.h
*
*   @brief  This file contains global definitions for circular buffer
*           manipulation.
*
*   These macros implement a circular buffer which employs get and put
*   pointers, in such a way that mutual exclusion is not required
*   (assumes one reader & one writer).
*
*   It requires that the circular buffer size be a power of two, and the
*   size of the buffer needs to smaller than the index. So an 8 bit index
*   supports a circular buffer upto ( 1 << 7 ) = 128 entries, and a 16 bit index
*   supports a circular buffer upto ( 1 << 15 ) = 32768 entries.
*
*   The basis for these routines came from an article in Jack Ganssle's
*   Embedded Muse: http://www.ganssle.com/tem/tem110.pdf
*
*   In order to offer the most amount of flexibility for embedded environments
*   you need to define macros for the size, entry type, and index type.
*
*   First, you need to name your circular buffer. For this example, we'll
*   call it myQ.
*
*   The size macro that needs to be defined will be the name of the
*   circular buffer followed by _SIZE. The size must be a power of two
*   and it needs to fit in the get/put indicies. i.e. if you use an
*   8 bit index, then the maximum supported size would be 128.
*
*   The entry type macro that needs to be defined will be named the name of the
*   circular buffer followed by _EntryType.
*
*   The index type macro that needs to be defined will be named the name of the
*   circular buffer followed by _indextype. its value should be an unsigned
*   integer type whose range is big enough to hold <queuename>_size entries.
*
*   Then you use the CBUF_DECLARE(name) macro to declare your queue.
*
*   For instance:
*
*       #define myQ_IndexType uint8_t
*       #define myQ_EntryType uint8_t
*       #define myQ_SIZE  64
*       CBUF_DECLARE(myQ);
*
*   You could then use
*
*       CBUF_Push( myQ, 'x' );
*
*   to add a character to the circular buffer, or
*
*       ch = CBUF_Pop( myQ );
*
*   to retrieve an element from the buffer.
*
*   If you happen to prefer to use C++ instead, there is a templatized
*   version which requires no macros. You just declare 3 template parameters:
*
*       - The type that should be used for the index
*       - The size of the circular buffer
*       - The type that should be used for the entry
*
*   For example:
*
*       CBUF< uint8_t, 64, char >   myQ;
*
****************************************************************************/

#if !defined( CBUF_H )
#define CBUF_H       /**< Include Guard                          */

#if !defined( __cplusplus )

/* ---- Constants and Types ---------------------------------------------- */

#define CBUF_DECLARE(name)                  \
struct  								    \
{                                           \
    volatile name##_IndexType m_getIdx;     \
    volatile name##_IndexType m_putIdx;     \
    name##_EntryType m_entry[ name##_SIZE ];\
} name;

/**
*   Initializes the circular buffer for use.
*/ 

#define CBUF_Init( cbuf )       cbuf.m_getIdx = cbuf.m_putIdx = 0

/**
*   Returns the number of elements which are currently contained in the 
 *  circular buffer.
*/

#define CBUF_Len( cbuf )        ((typeof( cbuf.m_putIdx ))(( cbuf.m_putIdx ) - ( cbuf.m_getIdx )))

/**
*   Appends an element to the end of the circular buffer
*/

#define CBUF_Push( cbuf, elem ) (cbuf.m_entry)[ cbuf.m_putIdx++ & (( cbuf##_SIZE ) - 1 )] = (elem)

/**
*   Returns a pointer to the last spot that was pushed.
*/

#define CBUF_GetLastEntryPtr( cbuf ) &(cbuf.m_entry)[ ( cbuf.m_putIdx - 1 ) & (( cbuf##_SIZE ) - 1 )]

/**
*   Returns a pointer to the next spot to push to.
*/

#define CBUF_GetPushEntryPtr( cbuf ) &(cbuf.m_entry)[ cbuf.m_putIdx & (( cbuf##_SIZE ) - 1 )]

/**
*   Advances the put index.
*/

#define CBUF_AdvancePushIdx( cbuf )  cbuf.m_putIdx++

/**
*   Retrieves an element from the beginning of the circular buffer
*/

#define CBUF_Pop( cbuf )        (cbuf.m_entry)[ cbuf.m_getIdx++ & (( cbuf##_SIZE ) - 1 )]

/**
*   Retrieves the i'th element from the beginning of the circular buffer
*/

#define CBUF_Get( cbuf, idx )        (cbuf.m_entry)[( cbuf.m_getIdx + idx ) & (( cbuf##_SIZE ) - 1 )]

/**
*   Retrieves the i'th element from the end of the circular buffer
*/

#define CBUF_GetEnd( cbuf, idx )        (cbuf.m_entry)[( cbuf.m_putIdx - idx - 1 ) & (( cbuf##_SIZE ) - 1 )]

/**
*   Returns a pointer to the next spot to pop from.
*/

#define CBUF_GetPopEntryPtr( cbuf ) &(cbuf.m_entry)[ cbuf.m_getIdx & (( cbuf##_SIZE ) - 1 )]

/**
*   Advances the get index.
*/

#define CBUF_AdvancePopIdx( cbuf )  cbuf.m_getIdx++

/**
*   Determines if the circular buffer is empty
*/

#define CBUF_IsEmpty( cbuf )    ( CBUF_Len( cbuf ) == 0 )

/**
*   Determines if the circular buffer is full.
*/

#define CBUF_IsFull( cbuf )     ( CBUF_Len( cbuf ) == ( cbuf##_SIZE ))

/**
*   Determines if the circular buffer is currenly overflowed or underflowed.
*/

#define CBUF_Error( cbuf )      ( CBUF_Len( cbuf ) > cbuf##_SIZE )

#else   /* defined(__cplusplus) */

template < class IndexType, IndexType Size, class EntryType >
class CBUF
{
	static const IndexType Mask = Size - 1;
    // Compile-time check for Size being power-of-2
    // If Size is not a power of 2, then Size&Mask will be non-0; hence
    // the array bounds will be -1, triggering a compile-time error.
    typedef EntryType SizeCheck[ !(Size & Mask) - 1 ];
    // keep index within range of [0..Size)
    static IndexType Wrapped(IndexType index) { return index & Mask; }

    // utilities; I don't think these should necessarily be public
	EntryType *GetLastEntryPtr()        { return m_entry + Wrapped( m_putIdx - 1 ); }
    EntryType *GetPushEntryPtr()        { return m_entry + Wrapped( m_putIdx ); }
    void AdvancePushIdx()               { m_putIdx = Wrapped( m_putIdx++ ); }

    EntryType *GetPopEntryPtr()         { return m_entry + Wrapped( m_getIdx ); }
    void AdvancePopIdx(IndexType skip=1){ m_getIdx = Wrapped( m_getIdx + skip ); }

public:

    // constructor (not needed if static)
//    CBUF() : m_getIdx(0), m_putIdx(0) { }

	// if not static, could call this to save a ctor.
	void Init() { m_putIdx = m_getIdx = 0; }

    // queries
	IndexType Len() const   { return m_putIdx - m_getIdx; }
    bool IsEmpty() const    { return Len() == 0; }
    bool IsFull() const     { return Len() == Size; }
    bool Error() const      { return Len() > Size; }

    // push next item on end (could leave buffer overflowed; check with Error()).
    void Push( EntryType val ) { m_entry[ Wrapped( m_putIdx++ ) ] = val; }

    // return next item from beginning (could leave buffer underflowed; check with Error()).
    EntryType Pop()         { return m_entry[ Wrapped( m_getIdx++ ) ]; }

    // skip n'th next items from beginning
    void Skip( IndexType skip=1 )   { AdvancePopIdx(skip); }

    // get idx'th element from beginning of buffer (0 returns next element to Pop())
    EntryType Get( IndexType idx )  { return m_entry[ Wrapped( m_getIdx + idx ) ]; }

    // get idx'th last element Pushed to buffer (0 returns last element Push()'d)
    EntryType GetEnd( IndexType idx ) { return m_entry[ Wrapped( m_putIdx - idx - 1 ) ]; }

private:

    volatile IndexType  m_getIdx;
    volatile IndexType  m_putIdx;
    EntryType           m_entry[ Size ];

};

#endif  // __cplusplus

/* ---- Variable Externs ------------------------------------------------- */
/* ---- Function Prototypes ---------------------------------------------- */

/** @} */

#endif // CBUF_H
